from pybdei import get_loglikelihood, ERRORS, WARNINGS, INFO, DEBUG, PYBDEI_VERSION, SAMPLING_PERIOD_LENGTH, MAX, MIN, \
    MEDIAN, MEAN


def main():
    """
    Entry point, calling :py:func:`pydbei.infer` with command-line arguments.
    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(description="BDEI model likelihood calculation on phylogenetic trees.", prog='bdei_loglikelihood')

    tree_group = parser.add_argument_group('tree-related arguments')
    tree_group.add_argument('--nwk', help="input tree(s) in newick format (must be rooted).",
                            type=str, required=True)
    tree_group.add_argument('-u', '--u', help="Number of unobserved trees. "
                                              "Specify -1 if you want u to be estimated.",
                            type=int, default=-1)

    parameter_group = parser.add_argument_group('parameter-related arguments')
    parameter_group.add_argument('--mu', required=True, type=float, default=None,
                                  help="Value to fix BDEI becoming-infectious rate mu. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--la', required=True, type=float, default=None,
                                  help="Value to fix BDEI transmission rate lambda. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--psi', required=True, type=float, default=None,
                                  help="Value to fix BDEI removal rate psi. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('-p', '--p', required=True, type=float, default=None,
                                  help="Value to fix BDEI sampling probability. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--pi_E', required=False, type=float, default=-1,
                                 help="Frequency of E at time 0, "
                                      "should be between 0 and 1. "
                                      "If not given, will be estimated from the model parameters.")
    parameter_group.add_argument('--T', default=0, type=float,
                                  help="Total time between the tree roots and the end of the sampling period. "
                                       "If a positive value is given, the total time will be set to the maximum "
                                       "between this value and the maximal time between the start "
                                       "and the last sampled tip of all the trees. "
                                       "If a zero or negative value is given, the time will be tree-specific "
                                       "and estimated as the time between the root "
                                       "and the last sampled tip for each tree. "
                                       "In the latter case, one can additionally annotate each tree root "
                                       "with a feature '{sp}' (e.g. '(a:2,b:3):1[&&NHX:{sp}=5];' "
                                       "is a tree with two tips, a and b, and the tree-specific time annotated to 5): "
                                       "then the tree-specific time will be set to the maximum "
                                       "between the annotated value and the time between the root "
                                       "and the last sampled tip of this tree.".format(sp=SAMPLING_PERIOD_LENGTH))
    parameter_group.add_argument('--u_policy', default=MEAN, choices=[MIN, MEDIAN, MEAN, MAX],
                                  help="How to estimate the time for unobserved trees "
                                       "in case of tree-specific observed tree times. "
                                       "By default, the mean of tree-specific observed times is taken.")

    parser.add_argument('-t', '--threads', help="number of threads for parallelization.", type=int, default=1)
    parser.add_argument('--log_level',
                        help="level of logging information "
                             "(the lower, the less information will be printed to the output). "
                             "Possible levels are: {} (errors only), {} (errors+warnings), {} (errors+warnings+info), "
                             "{} (errors+warnings+info+debug).".format(ERRORS, WARNINGS, INFO, DEBUG), type=int,
                        default=INFO)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {version}'.format(version=PYBDEI_VERSION))

    params = parser.parse_args()
    if params.pi_E is None:
        params.pi_E = -1
    res = get_loglikelihood(**vars(params))
    print(res)


if '__main__' == __name__:
    main()



import os
from collections import namedtuple

import _pybdei
import numpy as np
from ete3 import Tree

SAMPLING_PERIOD_LENGTH = 'T'

ERRORS = 0
WARNINGS = 1
INFO = 2
DEBUG = 3

PYBDEI_VERSION = 0.11

PS = (0.1, 0.4, 0.7)

MAX, MIN, MEDIAN, MEAN = 'max', 'min', 'median', 'mean'


BDEI_result = namedtuple('BDEI_result', ['mu', 'la', 'psi', 'p', 'mu_CI', 'la_CI', 'psi_CI', 'p_CI',
                                         'R_naught', 'incubation_period', 'infectious_time'])
BDEI_time = namedtuple('BDEI_time', ['CPU_time', 'iterations'])


def get_temp_file_name(prefix):
    return '{}.{}.temp'.format(prefix, np.random.random())


def parse_tree(nwk):
    """Parses an input tree in newick format and checks if it is binary and rooted"""
    tree = None
    for f in (0, 1, 2, 3, 5):
        try:
            tree = Tree(nwk, format=f)
            break
        except:
            pass
    if not tree:
        raise ValueError("Please make sure to use correct newick format and to specify branch lengths.")
    if len(tree.children) > 2:
        raise ValueError("The input tree must be rooted and binary, please root your tree.")
    for _ in tree.traverse():
        n_children = len(_.children)
        if not _.is_leaf() and n_children != 2:
            raise ValueError("The input tree must be binary, while your tree contains internal nodes with {} children, "
                             "please fix it.".format(n_children))
        if not _.is_root() and _.dist == 0:
            raise ValueError("The input tree must not contain non-root zero branches, while your tree does, "
                             "please fix it.")
    return tree


def parse_forest(nwk, T=0., u_policy=MEAN):
    forest = []
    with open(nwk, 'r') as f:
        forest = [parse_tree(_.strip('\n') + ';') for _ in f.read().split(';')[:-1]]
    if not forest:
        raise ValueError('Could not find any trees in your input file (check the newick format): {}'.format(nwk))
    Ts = [max(get_T(tree), float(getattr(tree, SAMPLING_PERIOD_LENGTH, 0))) for tree in forest]
    max_T = max(*Ts, T)
    for (tree, tree_T) in zip(forest, Ts):
        tree.add_feature(SAMPLING_PERIOD_LENGTH, tree_T if T <= 0 else max_T)
    u_T = max_T
    if T <= 0:
        if u_policy == MEDIAN:
            u_T = np.median(Ts)
        if u_policy == MEAN:
            u_T = np.mean(Ts)
        if u_policy == MIN:
            u_T = np.min(Ts)

    return forest, u_T


def save_forest(forest, nwk):
    with open(nwk, 'w+') as f:
        for tree in forest:
            _ = tree.write(format=5, format_root_node=True, features=[SAMPLING_PERIOD_LENGTH])\
                .replace('&&NHX:{}='.format(SAMPLING_PERIOD_LENGTH), '')
            f.write('{}\n'.format(_))


def get_T(tree):
    """
    Calculates the time between the tree start and the last sampled tip.
    """
    TIME = '_time'
    for n in tree.traverse('preorder'):
        if n.is_root():
            p_time = 0
        else:
            p_time = getattr(n.up, TIME)
            # delattr(n.up, TIME)
        n.add_feature(TIME, p_time + n.dist)
    time = 0
    for t in tree:
        time = max(time, getattr(t, TIME))
        # delattr(t, TIME)
    return time



def initial_rate_guess(forest, mu=None, la=None, psi=None):

    fixed_rates = []
    for rate in (mu, la, psi):
        if rate and rate > 0:
            fixed_rates.append(rate)
    if fixed_rates:
        avg_rate = np.average(fixed_rates)
        min_rate = avg_rate
        max_rate = avg_rate
    else:
        internal_is, external_is = [], []
        internal_es, external_es = [], []
        for tree in forest:
            for n in tree.traverse('preorder'):
                if not n.is_leaf():
                    i, e = sorted(n.children, key=lambda c: c.dist)
                    (internal_is if not i.is_leaf() else external_is).append(i)
                    (internal_es if not i.is_leaf() else external_es).append(e)
        i_s_times = np.array([_.dist for _ in external_is])
        i_tr_times = np.array([_.dist for _ in internal_is])
        e_s_times = np.array([_.dist for _ in external_es])
        e_tr_times = np.array([_.dist for _ in internal_es])
        i_s_time = np.median(i_s_times)
        e_s_time = np.median(e_s_times)
        # if it is a corner case when we only have tips, let's use sampling times
        i_tr_time = np.median(i_tr_times) if len(i_tr_times) else i_s_time
        e_tr_time = np.median(e_tr_times) if len(e_tr_times) else e_s_time
        mu_times = []
        if e_s_time > i_s_time:
            mu_times.append(e_s_time - i_s_time)
        if e_tr_time > i_tr_time:
            mu_times.append(e_tr_time - i_tr_time)
        mu_time = np.average(mu_times) if mu_times else min(i_s_time, i_tr_time) * 0.01
        avg_rate = np.average([1 / i_s_time, 1 / i_tr_time, 1 / mu_time])
        min_rate = min([1 / i_s_time, 1 / i_tr_time, 1 / mu_time])
        max_rate = max([1 / i_s_time, 1 / i_tr_time, 1 / mu_time])

    return sorted({avg_rate, min_rate, max_rate})


def infer(nwk, start=None, upper_bounds=None, pi_E=-1,
          mu=-1, la=-1, psi=-1, p=-1, T=0., u=-1, CI_repetitions=0, threads=1, log_level=INFO, u_policy=MEAN, **kwargs):
    """Infer BDEI parameters from a phylogenetic tree."""

    forest, u_T = parse_forest(nwk, T=T, u_policy=u_policy)

    if isinstance(start, BDEI_result):
        start = np.array([start.mu, start.la, start.psi, start.p])
    if start is None:
        if not p or p <= 0 or p >= 1:
            rate = initial_rate_guess(forest, mu, la, psi).pop()
            starts = [[rate, rate, rate, pp] for pp in PS]
        else:
            starts = [[rate, rate, rate, p] for rate in initial_rate_guess(forest, mu, la, psi)]
    else:
        starts = [start]
    for s in starts:
        s[-1] = min(0.99, max(0.001, s[-1]))

    if upper_bounds is None:
        upper_bounds = np.array([np.inf, np.inf, np.inf, 1])
    elif isinstance(upper_bounds, BDEI_result):
        upper_bounds = np.array([upper_bounds.mu, upper_bounds.la, upper_bounds.psi, upper_bounds.p])
    else:
        upper_bounds = np.array(upper_bounds)
    if any(upper_bounds <= 0):
        raise ValueError('Upper bound must be positive.')
    upper_bounds[-1] = min(upper_bounds[-1], 1)

    if pi_E is not None and pi_E >= 0:
        if pi_E > 1:
            raise ValueError('Frequency pi_E must be between 0 and 1.')
    else:
        pi_E = -1

    something_is_fixed = False
    all_is_fixed = True
    if mu >= 0:
        upper_bounds[0] = mu
        for s in starts:
            s[0] = mu
        something_is_fixed = True
    else:
        all_is_fixed = False
    if la >= 0:
        upper_bounds[1] = la
        for s in starts:
            s[1] = la
        something_is_fixed = True
    else:
        all_is_fixed = False
    if psi >= 0:
        upper_bounds[2] = psi
        for s in starts:
            s[2] = psi
        something_is_fixed = True
    else:
        all_is_fixed = False
    if 0 < p <= 1:
        upper_bounds[3] = p
        for s in starts:
            s[3] = p
        something_is_fixed = True
    else:
        all_is_fixed = False

    if not something_is_fixed:
        raise ValueError('To be identifiable, the BDEI model needs one of its parameters to be fixed. '
                         'Please do so, using one of the following arguments: mu, la, psi, p.')
    if all_is_fixed:
        raise ValueError('At least one of the following arguments: mu, la, psi, p, should be left to be optimised.')
    starts = [np.minimum(s, upper_bounds * 0.999) for s in starts]
    nstarts = len(starts)
    starts = np.reshape(starts, (4 * nstarts,))

    temp_nwk = get_temp_file_name(nwk)
    save_forest(forest, temp_nwk)
    res = _pybdei.infer(f=temp_nwk, start=starts, ub=upper_bounds, pie=pi_E,
                        mu=mu, la=la, psi=psi, p=p, u=u, ut=u_T, nt=threads, nbiter=CI_repetitions,
                        debug=log_level, nstarts=nstarts)
    try:
        os.remove(temp_nwk)
    except OSError:
        pass

    return BDEI_result(mu=res[0], la=res[1], psi=res[2], p=res[3],
                       mu_CI=(res[4], res[5]) if CI_repetitions > 0 else None,
                       la_CI=(res[6], res[7]) if CI_repetitions > 0 else None,
                       psi_CI=(res[8], res[9]) if CI_repetitions > 0 else None,
                       p_CI=(res[10], res[11]) if CI_repetitions > 0 else None,
                       R_naught=res[1] / res[2], incubation_period=1 / res[0], infectious_time=1 / res[2]), \
           BDEI_time(CPU_time=res[13], iterations=res[14])


def get_loglikelihood(nwk, mu=-1, la=-1, psi=-1, p=-1, pi_E=-1, T=0., u=-1, threads=1, log_level=INFO, params=None,
                      u_policy=MEAN, **kwargs):
    """Calculate loglikelihood for given BDEI parameters from a phylogenetic tree."""
    if params is not None:
        if isinstance(params, BDEI_result):
            mu = params.mu
            la = params.la
            psi = params.psi
            p = params.p
        else:
            [mu, la, psi, p] = params
    for par in [mu, la, psi, p]:
        if par is None or par < 0:
            raise ValueError('All the parameters (mu, la, psi, p) must be specified, '
                             'either via dedicated arguments or via the params argument')

    forest, u_T = parse_forest(nwk, T=T, u_policy=u_policy)

    temp_nwk = get_temp_file_name(nwk)
    save_forest(forest, temp_nwk)
    res = _pybdei.likelihood(f=temp_nwk, mu=mu, la=la, psi=psi, p=p, pie=pi_E, u=u, ut=u_T, nt=threads, debug=log_level)
    try:
        os.remove(temp_nwk)
    except OSError:
        pass

    return res




import numpy as np
from scipy.integrate import odeint
from treesimulator.mtbd_models import BirthDeathExposedInfectiousModel

from pybdei import parse_forest, PYBDEI_VERSION, ERRORS, WARNINGS, INFO, DEBUG, BDEI_result, get_T, \
    SAMPLING_PERIOD_LENGTH, MEDIAN, MEAN, MIN, MAX

RTOL = 100 * np.finfo(np.float64).eps
N_U_STEPS = int(1e7)


def compute_U(T, MU, LA, PSI, RHO, SIGMA=None, nsteps=N_U_STEPS):
    """
    Calculates a function get_U which for a given time t: 0 <= t <= T, would return
    an array of unobserved probabilities [U_1(t), ..., U_m(t)].

    U_k(t) are calculated by
    (1) solving their ODEs numerically for an array tt of nsteps times equally spaced between t=T and t=0,
    producing an array of solutions sol of length nstep (for corresponding times in tt)s.
    (2) creating a linear approximation which for a given time t (2a) find an index i such that tt[i] >= t > tt[i+1];
    (2b) returns sol[i + 1] + (sol[i] - sol[i + 1]) * (tt[i] - t) / (tt[i] - tt[i + 1]).


    :param T: time at end of the sampling period
    :param MU: an array of state transition rates
    :param LA: an array of transmission rates
    :param PSI: an array of removal rates
    :param RHO: an array of sampling probabilities
    :param SIGMA: an array of rate sums: MU.sum(axis=1) + LA.sum(axis=1) + PSI
    :return: a function that for a given time t returns the array of corresponding unsampled probabilities:
        t ->  [U_1(t), ..., U_m(t)].
    """

    if SIGMA is None:
        SIGMA = MU.sum(axis=1) + LA.sum(axis=1) + PSI

    tt = np.linspace(T, 0, nsteps)
    y0 = np.ones(LA.shape[0], np.float64)
    PSI_NOT_RHO = PSI * (1 - RHO)

    def pdf_U(U, t):
        dU = (SIGMA - LA.dot(U)) * U - MU.dot(U) - PSI_NOT_RHO
        return dU

    sol = odeint(pdf_U, y0, tt, rtol=RTOL)
    sol = np.maximum(sol, 0)
    return sol


def get_u(mu=-1, la=-1, psi=-1, p=-1, pi_E=-1, T=0.0, f=0, nwk=None, log_level=INFO, params=None, u_policy=MEAN,
          **kwargs):
    """Calculates u for given BDEI parameters and a given forest."""
    if params is not None:
        if isinstance(params, BDEI_result):
            mu = params.mu
            la = params.la
            psi = params.psi
            p = params.p
        else:
            [mu, la, psi, p] = params
    for par in [mu, la, psi, p]:
        if par is None or par < 0:
            raise ValueError('All the parameters (mu, la, psi, p) must be specified, '
                             'either via dedicated arguments or via the params argument')

    if pi_E is None:
        pi_E = -1

    if not nwk:
        if f is None or f <= 0 or T is None or T <= 0:
            raise ValueError('Either the forest file (via nwk argument), '
                             'or both the number of observed trees (f) and the total time (T) must be specified.')

    if nwk:
        forest, T = parse_forest(nwk, T, u_policy=u_policy)
        f = len(forest)
        if log_level >= INFO:
            print('The input forest contains {} trees.'.format(f))
    if log_level >= INFO:
        print('T is set to {}.'.format(T))

    model = BirthDeathExposedInfectiousModel(mu=mu, la=la, psi=psi, p=p)
    Us = compute_U(T, model.transition_rates, model.transmission_rates, model.removal_rates, model.ps)[-1]
    state_frequencies = np.array([pi_E, 1 - pi_E]) if 0 <= pi_E <= 1 else model.state_frequencies
    U = state_frequencies.dot(Us)
    u = f * U / (1 - U)

    if log_level >= INFO:
        print("The number of hidden trees u is estimated {}.".format(u))

    return u


def main():
    """
    Entry point, calling :py:func:`pydbei.u_calculator.get_u` with command-line arguments.
    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Calculator of the number of hidden trees in the forest under given BDEI model parameters.",
        prog='bdei_u')

    tree_group = parser.add_argument_group('tree-related arguments')
    tree_group.add_argument('--nwk', help="Observed tree(s) in newick format "
                                          "to calculate the total time T and the number of observed trees. "
                                          "If not given, --f and --T must be specified.",
                            type=str, required=False, default=None)
    tree_group.add_argument('--f', help="Number of observed trees. If not given, --nwk must be specified.",
                            type=int, required=False, default=0)

    parameter_group = parser.add_argument_group('parameter-related arguments')
    parameter_group.add_argument('--mu', required=True, type=float, default=None,
                                 help="Value to fix BDEI becoming-infectious rate mu. "
                                      "If not given, will be estimated.")
    parameter_group.add_argument('--la', required=True, type=float, default=None,
                                 help="Value to fix BDEI transmission rate lambda. "
                                      "If not given, will be estimated.")
    parameter_group.add_argument('--psi', required=True, type=float, default=None,
                                 help="Value to fix BDEI removal rate psi. "
                                      "If not given, will be estimated.")
    parameter_group.add_argument('-p', '--p', required=True, type=float, default=None,
                                 help="Value to fix BDEI sampling probability. "
                                      "If not given, will be estimated.")
    parameter_group.add_argument('--pi_E', required=False, type=float, default=-1,
                                 help="Frequency of E at time 0, "
                                      "should be between 0 and 1. "
                                      "If not given, will be estimated from the model parameters.")
    parameter_group.add_argument('--T', default=0, type=float,
                                 help="Total time between the hidden tree roots and the end of the sampling period. "
                                      "If the observed forest is not specified as --nwk, "
                                      "this time must be positive and will be taken as is."
                                      "If the observed forest is specified as --nwk and a positive --T value is given, "
                                      "the total time will be set to the maximum "
                                      "between this value and the maximal time between the start "
                                      "and the last sampled tip of all the trees. "
                                      "If a zero or negative --T value is given, "
                                      "the observed forest must be specified as --nwk: "
                                      "The time will be calculated based on the observed tree-specific times "
                                      "according to the measure specified in --u_policy. "
                                      "Observed tree-specific times are estimated as the time between the root "
                                      "and the last sampled tip for each tree. "
                                      "In the latter case, one can additionally annotate each tree root "
                                      "with a feature '{sp}' (e.g. '(a:2,b:3):1[&&NHX:{sp}=5];' "
                                      "is a tree with two tips, a and b, and the tree-specific time annotated to 5): "
                                      "then the tree-specific time will be set to the maximum "
                                      "between the annotated value and the time between the root "
                                      "and the last sampled tip of this tree.".format(sp=SAMPLING_PERIOD_LENGTH))
    parameter_group.add_argument('--u_policy', default=MEAN, choices=[MIN, MEDIAN, MEAN, MAX],
                                 help="How to estimate the time for unobserved trees "
                                      "in case of tree-specific observed tree times. "
                                      "By default, the mean of tree-specific observed times is taken.")
    parser.add_argument('--log_level',
                        help="level of logging information "
                             "(the lower, the less information will be printed to the output). "
                             "Possible levels are: {} (errors only), {} (errors+warnings), {} (errors+warnings+info), "
                             "{} (errors+warnings+info+debug).".format(ERRORS, WARNINGS, INFO, DEBUG), type=int,
                        default=INFO)
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s {version}'.format(version=PYBDEI_VERSION))

    params = parser.parse_args()
    u = get_u(**vars(params))
    print(u)


if '__main__' == __name__:
    main()



from pybdei import infer, ERRORS, DEBUG, WARNINGS, INFO, PYBDEI_VERSION, SAMPLING_PERIOD_LENGTH, MIN, MEDIAN, MEAN, MAX


def main():
    """
    Entry point, calling :py:func:`pydbei.infer` with command-line arguments.
    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(description="BDEI model parameter inference from phylogenetic trees.", prog='bdei_infer')

    tree_group = parser.add_argument_group('tree-related arguments')
    tree_group.add_argument('--nwk', help="Input tree(s) in newick format (must be rooted).",
                            type=str, required=True)
    tree_group.add_argument('-u', '--u', help="Number of unobserved trees. "
                                              "By default (-1) is estimated.",
                            type=int, default=-1)

    parameter_group = parser.add_argument_group('parameter-related arguments')
    parameter_group.add_argument('--mu', default=-1, type=float,
                                  help="Value to fix BDEI becoming-infectious rate mu. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--la', default=-1, type=float,
                                  help="Value to fix BDEI transmission rate lambda. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--psi', default=-1, type=float,
                                  help="Value to fix BDEI removal rate psi. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('-p', '--p', default=-1, type=float,
                                  help="Value to fix BDEI sampling probability. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--start', default=None, nargs=4, type=float,
                                  help="Starting values for parameter optimisation, "
                                       "should be 4 values in the following order: mu, lambda, psi, p. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--upper_bounds', default=None, nargs=4, type=float,
                                  help="Upper bound on parameter values for parameter optimisation, "
                                       "should be in the following order: mu, lambda, psi, p. "
                                       "If not given, will be estimated.")
    parameter_group.add_argument('--pi_E', default=-1, type=float,
                                  help="Frequency of E at time 0, "
                                       "should be between 0 and 1. "
                                       "If not given, will be estimated from the model parameters.")
    parameter_group.add_argument('--T', default=0, type=float,
                                  help="Total time between the tree roots and the end of the sampling period. "
                                       "If a positive value is given, the total time will be set to the maximum "
                                       "between this value and the maximal time between the start "
                                       "and the last sampled tip of all the trees. "
                                       "If a zero or negative value is given, the time will be tree-specific "
                                       "and estimated as the time between the root "
                                       "and the last sampled tip for each tree. "
                                       "In the latter case, one can additionally annotate each tree root "
                                       "with a feature '{sp}' (e.g. '(a:2,b:3):1[&&NHX:{sp}=5];' "
                                       "is a tree with two tips, a and b, and the tree-specific time annotated to 5): "
                                       "then the tree-specific time will be set to the maximum "
                                       "between the annotated value and the time between the root "
                                       "and the last sampled tip of this tree.".format(sp=SAMPLING_PERIOD_LENGTH))

    parameter_group.add_argument('--u_policy', default=MEAN, choices=[MIN, MEDIAN, MEAN, MAX],
                                  help="How to estimate the time for unobserved trees "
                                       "in case of tree-specific observed tree times. "
                                       "By default, the mean of tree-specific observed times is taken.")

    result_group = parser.add_argument_group('output-related arguments')
    result_group.add_argument('-c', '--CI_repetitions', default=0,
                              help="Number of repetitions for CI calculation "
                                   "(the higher, the more precise but also longer; a typical value is 100). "
                                   "If not specified, CIs will not be calculated.",
                              type=int)
    result_group.add_argument('--log', default=None, type=str,
                              help="Path to the output file where to write the estimates. "
                                   "If not given, the estimates will only be printed in the stdout")
    result_group.add_argument('--time_log', default=None, type=str,
                              help="Path to the output file where to write the time. "
                                   "If not given, the time will only be printed in the stdout")

    parser.add_argument('-t', '--threads', help="number of threads for parallelization.", type=int, default=1)
    parser.add_argument('--log_level',
                        help="level of logging information "
                             "(the lower, the less information will be printed to the output). "
                             "Possible levels are: {} (errors only), {} (errors+warnings), {} (errors+warnings+info), "
                             "{} (errors+warnings+info+debug).".format(ERRORS, WARNINGS, INFO, DEBUG), type=int,
                        default=INFO)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {version}'.format(version=PYBDEI_VERSION))

    params = parser.parse_args()
    res, time = infer(**vars(params))
    print(res)
    print(time)

    if params.log:
        with open(params.log, 'w+') as f:
            f.write('mu\tmu_CI\tla\tla_CI\tpsi\tpsi_CI\tp\tp_CI\tR_naught\tincubation_period\tinfectious_time\n'
                    .format('\tmu_CI' if params.CI_repetitions else '',
                            '\tla_CI' if params.CI_repetitions else '',
                            '\tpsi_CI' if params.CI_repetitions else '',
                            '\tp_CI' if params.CI_repetitions else ''))
            f.write('{mu}\t{mu_CI}\t{la}\t{la_CI}\t{psi}\t{psi_CI}\t{p}\t{p_CI}\t{R_naught}\t{incubation_period}\t{infectious_time}\n'
                    .format(**dict(zip(res._fields, res))))
    if params.time_log:
        with open(params.time_log, 'w+') as f:
            f.write('{}\n'.format('\t'.join(time._fields)))
            f.write('{}\n'.format('\t'.join('{}'.format(_) for _ in time)))


if '__main__' == __name__:
    main()
